# -*-coding: utf-8 -*-
# !/usr/bin/python

# 接收话题 --> 发布path话题，并保存位姿（TUM格式）

import rospy
import datetime
import time
import numpy as np
from threading import Thread
from ntrip_ros.msg import GnssPVTSolnMsg  # Copy from ublox_driver
from datetime import datetime, timedelta
from nav_msgs.msg import Path
from geometry_msgs.msg import PoseStamped
import math
from ntrip_ros.msg import my_gnss

LEAP_SECONDS = 18

def gnss_to_utc(gnss_week, gnss_tow):
    gps_epoch = datetime.datetime(1980, 1, 6, 0, 0, 0)
    delta_t = datetime.timedelta(seconds=gnss_week * 604800 + gnss_tow - 18)
    utc_epoch = datetime.datetime(1970, 1, 1, 0, 0, 0)
    utc_time = (gps_epoch + delta_t - utc_epoch).total_seconds()
    return utc_time


def degree2rad(degree):
    return degree * np.pi / 180.0


def BLH2xyz(BLH, rad=False):
    L = BLH[0]
    B = BLH[1]
    H = BLH[2]
    if not rad:
        L = degree2rad(L)
        B = degree2rad(B)
    a = 6378137.000000
    b = 6356752.314245
    e2 = 1 - (b / a) ** 2
    N = a / np.sqrt(1 - e2 * np.sin(B) ** 2)
    x = (N + H) * np.cos(B) * np.cos(L)
    y = (N + H) * np.cos(B) * np.sin(L)
    z = (N * (1 - e2) + H) * np.sin(B)
    return [x, y, z]


def ecef_to_enu(x, y, z, lat0, lon0, h0):
    a = 6378137.0  # WGS-84 semi-major axis
    b = 6356752.314245  # WGS-84 semi-minor axis
    f = (a - b) / a  # WGS-84 flattening factor

    # Compute the geodetic parameters of the reference point
    sinlat0 = math.sin(math.radians(lat0))
    coslat0 = math.cos(math.radians(lat0))
    sinlon0 = math.sin(math.radians(lon0))
    coslon0 = math.cos(math.radians(lon0))
    N0 = a / math.sqrt(1 - (2 * f - f * f) * sinlat0 * sinlat0)
    X0 = (N0 + h0) * coslat0 * coslon0
    Y0 = (N0 + h0) * coslat0 * sinlon0
    Z0 = ((1 - f) * (1 - f) * N0 + h0) * sinlat0

    # Compute the ECEF coordinates of the input point relative to the reference point
    # dx = x - X0
    # dy = y - Y0
    # dz = z - Z0

    dx = x
    dy = y
    dz = z

    # Rotate the ECEF coordinates to the local tangent plane (i.e., ENU coordinates)
    sinlat = math.sin(math.radians(lat0))
    coslat = math.cos(math.radians(lat0))
    sinlon = math.sin(math.radians(lon0))
    coslon = math.cos(math.radians(lon0))
    E = -sinlon * dx + coslon * dy
    N = -sinlat * coslon * dx - sinlat * sinlon * dy + coslat * dz
    U = coslat * coslon * dx + coslat * sinlon * dy + sinlat * dz

    return E, N, U

# a为绕Z轴逆时针旋转的角度
def rotate_z(x,y,z,a):
    a = math.radians(a)

    res_x = x*math.cos(a) - y*math.sin(a)
    res_y = x*math.sin(a) + y*math.cos(a)
    res_z = z

    return res_x, res_y, res_z

def call_back(msg, (data_list, path_msg, pub, xyz_first,LLA_first)):
    if msg is not None:
        data = ""
        print("******************************")
        BLH = []
        BLH.append(msg.longitude)  # 经度L
        BLH.append(msg.latitude)  # 纬度B
        BLH.append(msg.altitude)  # 椭球上方高度H
        xyz = BLH2xyz(BLH, False)

        gnss_time = msg.stamp

        path_msg.header.stamp = rospy.Time.now()
        path_msg.header.frame_id = "map"

        if xyz_first[-1] == 0:
            xyz_first[-1] = 1
            xyz_first[0] = xyz[0]
            xyz_first[1] = xyz[1]
            xyz_first[2] = xyz[2]

        if LLA_first[-1]== 0:
            LLA_first[-1] = 1
            LLA_first[0] = msg.latitude
            LLA_first[1] = msg.longitude
            LLA_first[2] = msg.altitude

        # ECEF的坐标 -> ECEF相对坐标
        x = xyz[0] - xyz_first[0]
        y = xyz[1] - xyz_first[1]
        z = xyz[2] - xyz_first[2]

        # ECEF的相对坐标 -> ENU坐标
        x,y,z = ecef_to_enu(x,y,z,LLA_first[0],LLA_first[1],LLA_first[2])
        # # 绕Z轴旋转
        # x,y,z = rotate_z(x,y,z,-80)

        k = 2
        x,y,z = k*x, k*y, k*z

        # 添加新的位姿，并发布出去
        pose1 = PoseStamped()
        pose1.header.frame_id = "map"
        pose1.header.stamp = rospy.Time.now()
        pose1.pose.position.x = x
        pose1.pose.position.y = y
        pose1.pose.position.z = z
        pose1.pose.orientation.w = 1.0

        path_msg.poses.append(pose1)
        pub.publish(path_msg)


        print("时间戳：{}\tXYZ坐标：{}".format(gnss_time,[x,y,z]))
        data += repr(gnss_time) + ' ' + repr(x) + ' ' + repr(y) + ' ' + repr(z) + ' ' + "0 0 0 1" + "\n"
        # data += repr(gnss_time) + ' ' + repr(xyz[0]) + ' ' + repr(xyz[1]) + ' ' + repr(xyz[2]) + ' ' + "0 0 0 1" + "\n"
        data_list.append(data)
    else:
        print("空数据")


def main():
    xyz_first = [0,0,0,0]  #[x,y,z,flag]
    LLA_first = [0,0,0,0] #[jing,wei,h,flag]
    data_list = []
    rospy.init_node("RTK", anonymous=False)
    print("节点初始化成功")

    pub = rospy.Publisher('RTK_path', Path, queue_size=10)
    path_msg = Path()
    print("发布【/RTK_path】成功")

    sub_gnss = rospy.Subscriber("/my_gnss_topic", my_gnss, call_back,
                                (data_list, path_msg, pub, xyz_first,LLA_first), queue_size=100)
    print("订阅【/ublox_driver/receiver_pvt】成功")

    rospy.spin()
    print("\n正在保存结果...")
    with open("result.txt", "w") as f:
        for one in data_list:
            f.write(one)
    print("保存完毕！保存至[./result.txt]")


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        print("end")
